import numpy as np
import pandas as pd
import margingame.payoff.costs as costs
from margingame.payoff.probabilities import gaussian_probabilities


def expected_payoff_with_costs(delta,
                               prob_b,
                               std_dev,
                               defender_naive_payoffs=(0, 25, -50),
                               adversary_naive_payoffs=(0, -10, 50),
                               delta_parameter=4,
                               prob_b_parameter=3,
                               std_dev_parameter=3,
                               defender=True):
    """
    :param delta: Radius of uncertainty margin.
    :param prob_b: Expected probability of attack being successful (breaching) given that there is no uncertainty
                        margin. In other words, the mean target of the attack.
    :param std_dev: Standard deviation in the result of the attack.
    :param defender_naive_payoffs: The defender's respective payoff of an attack failing, being detained and succeeding.
    :param adversary_naive_payoffs: The attacker's respective payoff of an attack failing, being detained and
                        succeeding.
    :param delta_parameter: Adjusts the cost of adjusting delta.
    :param prob_b_parameter: Adjusts the cost of adjusting prob_b.
    :param std_dev_parameter: Adjusts the cost of adjusting standard deviation.
    :param defender: If True, calculates defender's expected payoff. If False, calculates attacker's expected payoff.
    :return: Calculates a chosen player's expected payoff given the input parameters.
    """

    probabilities = gaussian_probabilities(delta, prob_b, std_dev)

    if defender:
        return np.dot(np.array(defender_naive_payoffs), probabilities) \
               - costs.cost_of_increasing_margin(delta, delta_parameter)
    else:
        return np.dot(np.array(adversary_naive_payoffs), probabilities) \
               - costs.cost_of_increasing_success_rate(prob_b, prob_b_parameter) \
               - costs.cost_of_reducing_std_dev(std_dev, std_dev_parameter)


def open_linspace(start, stop, step):
    """
    :param start: Start point.
    :param stop: End point.
    :param step: Number of elements desired.
    :return: List of 'step' evenly-spaced numbers between 'start' and 'end', excluding 'start' and 'end'.
    """
    return np.linspace(start, stop, step+2)[start+1:step+1]


def payoff_matrix_with_costs(no_of_deltas=4,
                             delta_limit=2,
                             no_of_probabilities=4,
                             no_of_std_devs=3,
                             std_dev_limit=2,
                             defender_naive_payoffs=(0, 25, -50),
                             adversary_naive_payoffs=(0, -10, 50),
                             delta_parameter=4,
                             prob_b_parameter=3,
                             std_dev_parameter=3,
                             defender=True):
    """
    :param no_of_deltas: Number of evenly spaced deltas (size of uncertainty margin) to be investigated in the
                         range (0, 'delta_limit').
    :param delta_limit: The exclusive endpoint that the deltas go up to.
    :param no_of_probabilities: Number of evenly spaced probabilities (attacker targets) to be investigated in the
                                range (0, 1).
    :param no_of_std_devs: Number of evenly spaced attacker standard deviations to be investigated in the
                            range (0, 'std_dev_limit').
    :param std_dev_limit: The exclusive endpoint that the standard deviations go up to.
    :param defender_naive_payoffs: The defender's respective payoff of an attack failing, being detained and succeeding.
    :param adversary_naive_payoffs: The attacker's respective payoff of an attack failing, being detained and
                        succeeding.
    :param delta_parameter: Adjusts the cost of adjusting delta.
    :param prob_b_parameter: Adjusts the cost of adjusting prob_b.
    :param std_dev_parameter: Adjusts the cost of adjusting standard deviation.
    :param defender: If True, calculates defender's expected payoff matrix. If False, calculates attacker's expected
                     payoff matrix.
    :return: Calculates a chosen player's expected payoff matrix given the input parameters.
    """
    no_of_columns = no_of_probabilities * no_of_std_devs
    no_of_rows = no_of_deltas

    list_of_probabilities = [x.round(1) for x in open_linspace(0, 1, no_of_probabilities)]
    list_of_std_devs = [x.round(1) for x in open_linspace(0, std_dev_limit, no_of_std_devs)]
    columns = pd.MultiIndex.from_product([list_of_probabilities, list_of_std_devs], names=['Target', 'Standard Deviation'])

    list_of_deltas = [x.round(1) for x in open_linspace(0, delta_limit, no_of_deltas)]

    payoff_matrix = pd.DataFrame(np.zeros((no_of_rows, no_of_columns)), index=list_of_deltas, columns=columns)

    for row in range(0, no_of_rows):
        for column in range(0, no_of_columns):
            payoff_matrix.iloc[row, column] = expected_payoff_with_costs(list_of_deltas[row],
                                                                         columns[column][0],
                                                                         std_dev=columns[column][1],
                                                                         adversary_naive_payoffs=adversary_naive_payoffs,
                                                                         defender_naive_payoffs=defender_naive_payoffs,
                                                                         delta_parameter=delta_parameter,
                                                                         prob_b_parameter=prob_b_parameter,
                                                                         std_dev_parameter=std_dev_parameter,
                                                                         defender=defender)

    return payoff_matrix



import numpy as np


def cost_of_increasing_margin(delta, scaling_parameter=1):
    """
    :param delta: Radius of uncertainty margin.
    :param scaling_parameter: Adjusts the cost.
    :return: Cost of setting 'delta'.
    """
    return scaling_parameter * (np.exp(delta))


def cost_of_increasing_success_rate(prob_b, scaling_parameter=1):
    """
    :param prob_b: Expected probability of attack being successful (breaching) given that there is no uncertainty
                        margin. In other words, the mean target of the attack.
    :param scaling_parameter: Adjusts the cost.
    :return: Cost of setting 'prob_b'.
    """
    return scaling_parameter * np.tan((np.pi / 2) * prob_b)


def cost_of_reducing_std_dev(std_dev=1, scaling_parameter=0.2):
    """
    :param std_dev: Standard deviation in the result of the attack.
    :param scaling_parameter: Adjusts the cost.
    :return: Cost of setting 'std_dev'.
    """
    return scaling_parameter / std_dev



import numpy as np
from scipy.stats import norm


def gaussian_probabilities(delta, prob_b_without_margin, std_dev=1):
    """
    :param delta: Radius of uncertainty margin.
    :param prob_b_without_margin: Expected probability of attack being successful (breaching) given that there is no
                                    uncertainty margin. In other words, the mean target of the attack.
    :param std_dev: Standard deviation in the result of the attack.
    :returns: Numpy array containing the probability of attack failing, being detained and succeeding, respectively.
    """

    midpoint = norm.ppf(1 - prob_b_without_margin)  # Finds the midpoint of the margin.

    prob_q = norm.cdf(midpoint - delta, scale=std_dev)  # Gives area to the left of the left side of the margin.

    prob_b = 1 - norm.cdf(midpoint + delta, scale=std_dev)  # Gives area to the right of the right side
    # of the margin.

    prob_d = 1 - prob_b - prob_q  # Gives area in the margin

    return np.array([prob_q, prob_d, prob_b])



import nashpy as nash
import dask.array as da
from margingame.payoff.payoffs import payoff_matrix_with_costs

#  https://nashpy.readthedocs.io/en/stable/tutorial/index.html#creating-a-game


class Initialise:
    def __init__(self,
                 no_of_deltas=10,
                 delta_limit=2,
                 no_of_probabilities=4,
                 no_of_std_devs=4,
                 std_dev_limit=2,
                 defender_naive_payoffs=(0, 25, -50),
                 adversary_naive_payoffs=(0, -10, 50),
                 delta_parameter=5,
                 prob_b_parameter=5,
                 std_dev_parameter=1):

        self.attacker_payoff_matrix = payoff_matrix_with_costs(no_of_deltas=no_of_deltas,
                                                               delta_limit=delta_limit,
                                                               no_of_probabilities=no_of_probabilities,
                                                               no_of_std_devs=no_of_std_devs,
                                                               std_dev_limit=std_dev_limit,
                                                               adversary_naive_payoffs=adversary_naive_payoffs,
                                                               defender_naive_payoffs=defender_naive_payoffs,
                                                               delta_parameter=delta_parameter,
                                                               prob_b_parameter=prob_b_parameter,
                                                               std_dev_parameter=std_dev_parameter,
                                                               defender=False)

        self.defender_payoff_matrix = payoff_matrix_with_costs(no_of_deltas=no_of_deltas,
                                                               delta_limit=delta_limit,
                                                               no_of_probabilities=no_of_probabilities,
                                                               no_of_std_devs=no_of_std_devs,
                                                               std_dev_limit=std_dev_limit,
                                                               adversary_naive_payoffs=adversary_naive_payoffs,
                                                               defender_naive_payoffs=defender_naive_payoffs,
                                                               delta_parameter=delta_parameter,
                                                               prob_b_parameter=prob_b_parameter,
                                                               std_dev_parameter=std_dev_parameter,
                                                               defender=True)

        self.A = self.attacker_payoff_matrix.to_numpy()
        self.B = self.defender_payoff_matrix.to_numpy()

        no_of_columns = no_of_probabilities * no_of_std_devs
        no_of_rows = no_of_deltas
        self.dask_A = da.from_array(self.A, (no_of_rows, no_of_columns))
        self.dask_B = da.from_array(self.B, (no_of_rows, no_of_columns))

    def calculate_equilibria_support_enum(self, A=None, B=None):
        """
        :return: List of equilibria for the game found via support enumeration
        """
        if A is None and B is None:
            A, B = self.dask_A, self.dask_B

        game = nash.Game(A, B)
        equilibria = game.support_enumeration()
        return list(equilibria)

    def calculate_equilibria_vertex_enum(self, A=None, B=None):
        """
        :return: List of equilibria for the game found via support enumeration
        """
        if A is None and B is None:
            A, B = self.dask_A, self.dask_B

        game = nash.Game(A, B)
        equilibria = game.vertex_enumeration()
        return list(equilibria)
